@@ -1,54 +1,60 @@
 """Transformers that use stanza documents."""
+from __future__ import annotations
 
 import json
 import logging
 import os
 import warnings
+from typing import Any, Dict, Iterable, List, Union
 
 import nltk
-import stanza
 from sklearn.base import BaseEstimator, TransformerMixin
-from stanza import Document
-from stanza.models.common.doc import Sentence
 from tqdm import tqdm
 
+import stanza
+from stanza.models.common.doc import Document, Sentence, Word
+
 logger = logging.getLogger(__name__)
 
-UNKNOWN_KEY = 'UNK'
-WORD_FEATURES_JSON = os.path.join('resources', 'word_features.json')
+UNKNOWN_KEY = "UNK"
+WORD_FEATURES_JSON = os.path.join("resources", "word_features.json")
 
 
-def _get_word_attribute(word, attribute, fallback='_'):
+def _get_word_attribute(
+    word: Word,
+    attribute: str,
+    fallback: str = "_",
+) -> Union[str]:
     if fallback is None:
         raise ValueError("Fallback can't be None")
     if hasattr(word, attribute):
         a = getattr(word, attribute)
         if a is not None:
             return a
-    if not hasattr(word, '_feats'):
+    if not hasattr(word, "_feats"):
         return fallback
 
     feats = word.feats
-    if not feats or feats == '_':
+    if not feats or feats == "_":
         return fallback
-    pairs = feats.split('|')
-    features = {k: v for (k, v) in [x.split('=') for x in pairs]}
+    pairs = feats.split("|")
+    features = {k: v for (k, v) in [x.split("=") for x in pairs]}
     if attribute not in features or features[attribute] is None:
         return fallback
     return features[attribute]
 
 
-def _get_features_per_word(document):
+def _get_features_per_word(document: Document) -> List[Dict[str, str]]:
     result = []
     words = []
     for sentence in document.sentences:
         words += sentence.words
     for word in words:
         features = word.feats
-        if not features or features == '_':
+        if not features or features == "_":
             continue
-        pairs = features.split('|')
-        result.append({k: v for (k, v) in [x.split('=') for x in pairs]})
+        pairs = features.split("|")
+        result.append({k: v for (k, v) in [x.split("=") for x in pairs]})
     return result
 
 
@@ -63,30 +69,34 @@ class StanzaWordFeatureFrequencyTransformer(BaseEstimator, TransformerMixin):
     If this transformer encounters an element not in that file, it will add one
     """
 
-    def __init__(self):
+    def __init__(self) -> None:
         """Initialize class."""
+        super().__init__()
         with open(WORD_FEATURES_JSON) as i_fh:
             self.features = json.load(i_fh)
             self.features.append(UNKNOWN_KEY)  # for unknown keys
 
-    def fit(self, X, y=None, **fit_kwargs):
+    def fit(
+        self, _x: Any, _y: Any = None, **_fit_kwargs: Any
+    ) -> StanzaWordFeatureFrequencyTransformer:
         """Fit the model."""
         return self
 
-    def transform(self, X, y=None):
+    def transform(
+        self, x: Iterable[Document], _y: Iterable = None
+    ) -> List[Dict[str, int]]:
         """Transform documents."""
         result = []
-        for document in X:
+        for document in x:
             doc = {}
             for feature in self.features:
                 doc[feature] = 0
             for features_dict in _get_features_per_word(document):
                 for k, v in features_dict.items():
-                    key = f'{k}__{v}'
+                    key = f"{k}__{v}"
                     if key not in doc:
                         key = UNKNOWN_KEY
-                        warnings.warn(
-                            f'key {key} not found during fit, returning UNK')
+                        warnings.warn(f"key {key} not found during fit, returning UNK")
                     doc[key] += 1
             result.append(doc)
         return result
@@ -103,7 +113,7 @@ class StanzaNlpToFieldTransformer(BaseEstimator, TransformerMixin):
         Sentences are separated by newlines.
     """
 
-    def __init__(self, field=None):
+    def __init__(self, field: str):
         """
         Initialize class.
 
@@ -112,29 +122,31 @@ class StanzaNlpToFieldTransformer(BaseEstimator, TransformerMixin):
         """
         self.field = field
 
-    def fit(self, x, y=None):
+    def fit(self, _x: Any, _y: Any = None) -> StanzaNlpToFieldTransformer:
         """Fit the model."""
         return self
 
-    def transform(self, x, y=None):
+    def transform(
+        self, x: Iterable[Union[str, Document, Sentence]], _y: Any = None
+    ) -> List[str]:
         """Transform documents."""
         result = []
         for document in x:
-            document_result = []
-            sentences: [Sentence] = []
-            if type(document) == Sentence:
+            document_result: List[str] = []
+            sentences: List[Sentence] = []
+            if isinstance(document, Sentence):
                 sentences = [document]
-            elif type(document) == Document:
+            elif isinstance(document, Document):
                 sentences = document.sentences
             for sentence in sentences:
                 new_sentence = []
                 for word in sentence.words:
-                    label = _get_word_attribute(word, self.field, fallback='_')
+                    label = _get_word_attribute(word, self.field, fallback="_")
                     if label is None:
-                        raise ValueError(f'label is None for word: {word}')
+                        raise ValueError(f"label is None for word: {word}")
                     new_sentence.append(label)
-                document_result.append(' '.join(new_sentence))
-            result.append('\n'.join(document_result))
+                document_result.append(" ".join(new_sentence))
+            result.append("\n".join(document_result))
         return result
 
 
@@ -147,7 +159,7 @@ class StanzaParserTransformer(BaseEstimator, TransformerMixin):
     features from natural text.
     """
 
-    def __init__(self, language, silent=False, cpu=False):
+    def __init__(self, language: str, silent: bool = False, cpu: bool = False):
         """
         Initialize class.
 
@@ -161,15 +173,15 @@ class StanzaParserTransformer(BaseEstimator, TransformerMixin):
         self.silent = silent
         self.cpu = cpu
         if cpu:
-            logger.info('using CPU for stanford parsing')
-            os.environ['CUDA_VISIBLE_DEVICES'] = ''
+            logger.info("using CPU for stanford parsing")
+            os.environ["CUDA_VISIBLE_DEVICES"] = ""
         self.nlp = stanza.Pipeline(lang=self.language, use_gpu=not cpu)
 
-    def fit(self, x, y=None):
+    def fit(self, _x: Any, _y: Any = None) -> StanzaParserTransformer:
         """Fit the model."""
         return self
 
-    def transform(self, x, y=None):
+    def transform(self, x: Iterable[str], _y: Any = None) -> List[Document]:
         """Transform documents."""
         result = []
         if self.silent:
@@ -177,8 +189,8 @@ class StanzaParserTransformer(BaseEstimator, TransformerMixin):
         else:
             pbar = tqdm(x)
         for document in pbar:
-            with open('/tmp/stanford_parsing', 'a') as log_fh:
-                log_fh.write(document + '/n')
+            with open("/tmp/stanford_parsing", "a") as log_fh:
+                log_fh.write(document + "/n")
                 result.append(self.nlp(document))
         return result
 
@@ -197,7 +209,9 @@ class StanzaToNltkTreesTransformer(BaseEstimator, TransformerMixin):
 
     """
 
-    def __init__(self, node_labels=None):
+    node_labels: List[str]
+
+    def __init__(self, node_labels: Union[str, List[str]] = None):
         """
         Initialize transformer.
 
@@ -206,24 +220,27 @@ class StanzaToNltkTreesTransformer(BaseEstimator, TransformerMixin):
                 list of ['lemma', 'upos', 'xpos', 'dependency'], e.g.:
                 node_contents=['upos', 'dependency']
         """
-        if type(node_labels) == str:
-            node_labels = [node_labels]
+        if isinstance(node_labels, str):
+            self.node_labels = [node_labels]
+        elif isinstance(node_labels, list):
+            self.node_labels = node_labels
         if not node_labels:
-            node_labels = ['upos']
-        self.node_labels = node_labels
+            self.node_labels = ["upos"]
 
-    def fit(self, x, y=None):
+    def fit(self, _x: Any, _y: Any = None) -> StanzaToNltkTreesTransformer:
         """Fit the model."""
         return self
 
-    def transform(self, x, y=None):
+    def transform(
+        self, x: Iterable[Union[Document, Sentence]], _y: Any = None
+    ) -> List[List[nltk.Tree]]:
         """Transform documents."""
         result = []
         for document in x:
             sentences = []
-            if type(document) == Document:
+            if isinstance(document, Document):
                 sentences = document.sentences
-            elif type(document) == Sentence:
+            elif isinstance(document, Sentence):
                 sentences = [document]
             doc = []
             for sentence in sentences:
@@ -231,12 +248,11 @@ class StanzaToNltkTreesTransformer(BaseEstimator, TransformerMixin):
                 labeled_tree = self.get_symbols(dependency_tree)
                 doc.append(labeled_tree)
             if not doc:
-                raise ValueError(
-                    f'coud not parse anything from input {document}')
+                raise ValueError(f"coud not parse anything from input {document}")
             result.append(doc)
         return result
 
-    def get_symbols(self, tree):
+    def get_symbols(self, tree: nltk.Tree) -> nltk.Tree:
         """Extract the symbol from the dependency relationship."""
         dependency = tree.label()
         if isinstance(dependency, str):
@@ -247,16 +263,18 @@ class StanzaToNltkTreesTransformer(BaseEstimator, TransformerMixin):
         _, relationship, word_2 = dependency
         # the first word is already handled in other iterations of this method
         labels = []
-        if 'dependency' in self.node_labels:
+        if "dependency" in self.node_labels:
             labels.append(relationship)
-        labels += [_get_word_attribute(word_2, label, fallback='_')
-                   for label in self.node_labels
-                   if label != 'dependency']
-        label = '#'.join(labels)
+        labels += [
+            _get_word_attribute(word_2, label, fallback="_")
+            for label in self.node_labels
+            if label != "dependency"
+        ]
+        label = "#".join(labels)
         result = nltk.Tree(label, [self.get_symbols(c) for c in tree])
         return result
 
-    def get_nodes(self, tree):
+    def get_nodes(self, tree: nltk.Tree) -> List[nltk.Tree]:
         """Extract the leaf nodes from a tree."""
         leaves = []
         if isinstance(tree, nltk.Tree):
@@ -266,12 +284,12 @@ class StanzaToNltkTreesTransformer(BaseEstimator, TransformerMixin):
                 leaves += self.get_nodes(child)
         return leaves
 
-    def parse(self, sentence):
+    def parse(self, sentence: Sentence) -> nltk.Tree:
         """Construct a NLTK tree from stanza dependencies."""
-        roots = [d for d in sentence.dependencies if d[1] == 'root']
+        roots = [d for d in sentence.dependencies if d[1] == "root"]
         unprocessed = [d for d in sentence.dependencies if d not in roots]
         # it is possible that multiple words have a root relationship
-        tree = nltk.Tree('root', [nltk.Tree(x, []) for x in roots])
+        tree = nltk.Tree("root", [nltk.Tree(x, []) for x in roots])
 
         while True:
             changed = False
@@ -279,18 +297,17 @@ class StanzaToNltkTreesTransformer(BaseEstimator, TransformerMixin):
             nodes = self.get_nodes(tree)
             for d in unprocessed:
                 for leaf in nodes:
-                    if hasattr(d[0], 'id'):
+                    if hasattr(d[0], "id"):
                         if leaf.label()[2].id == d[0].id:
                             leaf.append(nltk.Tree(d, []))
                             added_dependencies.append(d)
                             changed = True
-                    elif hasattr(d[0], 'index'):
+                    elif hasattr(d[0], "index"):
                         if leaf.label()[2].index == d[0].index:
                             leaf.append(nltk.Tree(d, []))
                             added_dependencies.append(d)
                             changed = True
-            unprocessed = [x for x in unprocessed
-                           if x not in added_dependencies]
+            unprocessed = [x for x in unprocessed if x not in added_dependencies]
             if len(unprocessed) == 0 or not changed:
                 break
 
